@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2017. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -32,8 +32,8 @@
 %%                         (Optional - default efile)
 %%        -hosts [Node]  : List of hosts from which we can boot.
 %%                         (Mandatory if -loader inet)
-%%        -mode embedded : Load all modules at startup, no automatic loading
-%%        -mode interactive : Auto load modules (default system behaviour).
+%%        -mode interactive : Auto load modules not needed at startup (default system behaviour).
+%%        -mode embedded : Load all modules in the boot script, disable auto loading.
 %%        -path          : Override path in bootfile.
 %%        -pa Path+      : Add my own paths first.
 %%        -pz Path+      : Add my own paths last.
@@ -200,10 +200,11 @@ boot(BootArgs) ->
     register(init, self()),
     process_flag(trap_exit, true),
 
-    %% Load the zlib nif
+    %% Load the static nifs
     zlib:on_load(),
-    %% Load the tracer nif
     erl_tracer:on_load(),
+    prim_buffer:on_load(),
+    prim_file:on_load(),
 
     {Start0,Flags,Args} = parse_boot_args(BootArgs),
     %% We don't get to profile parsing of BootArgs
@@ -484,7 +485,12 @@ do_handle_msg(Msg,State) ->
 	X ->
 	    case whereis(user) of
 		undefined ->
-		    catch error_logger ! {info, self(), {self(), X, []}};
+		    Time = erlang:monotonic_time(microsecond),
+                    catch logger ! {log, info, "init got unexpected: ~p", [X],
+                                    #{pid=>self(),
+                                      gl=>self(),
+                                      time=>Time,
+                                      error_logger=>#{tag=>info_msg}}};
 		User ->
 		    User ! X,
 		    ok
@@ -544,6 +550,8 @@ stop(Reason,State) ->
     do_stop(Reason,State1).
 
 do_stop(restart,#state{start = Start, flags = Flags, args = Args}) ->
+    %% Make sure we don't have any outstanding messages before doing the restart.
+    flush(),
     boot(Start,Flags,Args);
 do_stop(reboot,_) ->
     halt();
@@ -556,8 +564,18 @@ do_stop({stop,Status},State) ->
 
 clear_system(BootPid,State) ->
     Heart = get_heart(State#state.kernel),
-    shutdown_pids(Heart,BootPid,State),
-    unload(Heart).
+    Logger = get_logger(State#state.kernel),
+    shutdown_pids(Heart,Logger,BootPid,State),
+    unload(Heart),
+    kill_em([Logger]),
+    do_unload([logger_server]).
+
+flush() ->
+    receive
+        _M -> flush()
+    after 0 ->
+            ok
+    end.
 
 stop_heart(State) ->
     case get_heart(State#state.kernel) of
@@ -570,19 +588,26 @@ stop_heart(State) ->
 	    shutdown_kernel_pid(Pid, BootPid, self(), State) 
     end.
 
-shutdown_pids(Heart,BootPid,State) ->
+shutdown_pids(Heart,Logger,BootPid,State) ->
     Timer = shutdown_timer(State#state.flags),
     catch shutdown(State#state.kernel,BootPid,Timer,State),
-    kill_all_pids(Heart), % Even the shutdown timer.
-    kill_all_ports(Heart),
+    kill_all_pids(Heart,Logger), % Even the shutdown timer.
+    kill_all_ports(Heart), % Logger has no ports
     flush_timout(Timer).
 
-get_heart([{heart,Pid}|_Kernel]) -> Pid;
-get_heart([_|Kernel])           -> get_heart(Kernel);
-get_heart(_)                    -> false.
+get_heart(Kernel) ->
+    get_kernelpid(heart,Kernel).
+
+get_logger(Kernel) ->
+    get_kernelpid(logger,Kernel).
+
+get_kernelpid(Name,[{Name,Pid}|_Kernel]) -> Pid;
+get_kernelpid(Name,[_|Kernel])           -> get_kernelpid(Name,Kernel);
+get_kernelpid(_,_)                    -> false.
 
 
-shutdown([{heart,_Pid}|Kernel],BootPid,Timer,State) ->
+shutdown([{Except,_Pid}|Kernel],BootPid,Timer,State)
+  when Except==heart; Except==logger ->
     shutdown(Kernel, BootPid, Timer, State);
 shutdown([{_Name,Pid}|Kernel],BootPid,Timer,State) ->
     shutdown_kernel_pid(Pid, BootPid, Timer, State),
@@ -634,24 +659,25 @@ resend(_) ->
 
 %%
 %% Kill all existing pids in the system (except init and heart).
-kill_all_pids(Heart) ->
-    case get_pids(Heart) of
+kill_all_pids(Heart,Logger) ->
+    case get_pids(Heart,Logger) of
 	[] ->
 	    ok;
 	Pids ->
 	    kill_em(Pids),
-	    kill_all_pids(Heart)  % Continue until all are really killed.
+	    kill_all_pids(Heart,Logger)  % Continue until all are really killed.
     end.
     
 %% All except system processes.
-get_pids(Heart) ->
+get_pids(Heart,Logger) ->
     Pids = [P || P <- processes(), not erts_internal:is_system_process(P)],
-    delete(Heart,self(),Pids).
+    delete(Heart,Logger,self(),Pids).
 
-delete(Heart,Init,[Heart|Pids]) -> delete(Heart,Init,Pids);
-delete(Heart,Init,[Init|Pids])  -> delete(Heart,Init,Pids);
-delete(Heart,Init,[Pid|Pids])   -> [Pid|delete(Heart,Init,Pids)];
-delete(_,_,[])                  -> [].
+delete(Heart,Logger,Init,[Heart|Pids]) -> delete(Heart,Logger,Init,Pids);
+delete(Heart,Logger,Init,[Logger|Pids])  -> delete(Heart,Logger,Init,Pids);
+delete(Heart,Logger,Init,[Init|Pids])  -> delete(Heart,Logger,Init,Pids);
+delete(Heart,Logger,Init,[Pid|Pids])   -> [Pid|delete(Heart,Logger,Init,Pids)];
+delete(_,_,_,[])                  -> [].
     
 kill_em([Pid|Pids]) ->
     exit(Pid,kill),
@@ -681,9 +707,9 @@ kill_all_ports(_,_) ->
     ok.
 
 unload(false) ->
-    do_unload(sub(erlang:pre_loaded(),erlang:loaded()));
+    do_unload(sub([logger_server|erlang:pre_loaded()],erlang:loaded()));
 unload(_) ->
-    do_unload(sub([heart|erlang:pre_loaded()],erlang:loaded())).
+    do_unload(sub([heart,logger_server|erlang:pre_loaded()],erlang:loaded())).
 
 do_unload([M|Mods]) ->
     catch erlang:purge_module(M),
